import matplotlib.pyplot as plt
import seaborn as sns
from scipy.cluster import hierarchy as hira
from scipy.spatial.distance import squareform
from scipy.cluster.hierarchy import linkage
from sklearn.metrics import precision_score
import pandas as pd

def dendro(similarity, labels, distancecut = 1):

    # HEATMAP
    f=plt.figure(figsize=(16,8))
    ax=plt.subplot(121)
    ax.set_title('complete similarity matrix', fontsize=20)
    sns.heatmap(similarity, xticklabels = False,
            yticklabels = labels,  cmap="YlGnBu")
    locs, labels = plt.yticks()
    plt.setp(labels,size = 8)

    # DENDROGRAM
    ax=plt.subplot(122)
    ax.set_title('induced dendrogram (ward)', fontsize=20)
    Z = squareform(similarity)
    Z = hira.linkage(1-Z,'ward')
    hira.dendrogram(Z, labels = labels, color_threshold=distancecut, orientation='right')
    locs, labels = plt.xticks()
    plt.setp(labels, rotation=90,size = 8)
    plt.subplots_adjust(wspace = .4)
    return hira.fcluster(Z, t=7, criterion = 'maxclust')



def plotPrecision(data, wg = True, method='cosine similarity'):
    # plor precision at k :)
    # [val,rep,p@k,n_genes]
    df = pd.DataFrame(data)
    df.columns = 'precision rep neighbors±σ genes '.split()
    title = f'Searching for similar datasets via {method}'

    if wg:
        sns.set_theme(style='whitegrid')
    sns.lineplot(data = df,x='genes', y = 'precision',style= 'neighbors±σ',
            hue = 'neighbors±σ',palette="flare", ci=68)
    plt.title(title, y=1.06, fontsize = 16)
    plt.ylim([.75,1])
    plt.ylabel('precision of neighbors (40 datasets)')
    plt.xlabel('number of genes')








from sklearn.metrics import adjusted_rand_score
import numpy as np
def evaluate(pairs, labels, setid = 1):
    scores = np.array([scorepair(p,l,setid = setid) for p,l in zip(pairs, labels)])
    return scores

def scorepair(pair, labels, setid=1):
    return adjusted_rand_score( pair.data[setid].obs['true'],labels[setid])



from lmz import Map,Zip,Filter,Grouper,Range,Transpose, grouper
import numpy as np
from ubergauss import tools as ut
from sklearn.metrics.pairwise import cosine_similarity as cos
from sklearn import preprocessing
from sklearn.metrics import precision_score


def cosine(a,b, numgenes = 500):
    scr1, scr2 = a.varm['scores'], b.varm['scores']
    if numgenes:
        mask = scr1+scr2
        mask = ut.binarize(mask,numgenes).astype(np.bool)
        scr1 = scr1[mask]
        scr2 = scr2[mask]
    else:
        assert False
    return cos([scr1],[scr2]).item()



def jaccard(a,b ,ngenes= False):
    # intersect/union
    assert n_genes > 3
    asd = np.array([ ut.binarize(d.varm['scores'],ngenes)  for d in [a,b]])
    union = np.sum(np.any(asd, axis=0))
    intersect = np.sum(np.sum(asd, axis=0) ==2)
    return intersect/union



def precision(matrix,shortlabels, k):
    truth = preprocessing.LabelEncoder().fit_transform(shortlabels)
    srt = np.argsort(matrix, axis=1)
    #print(f'{truth=} {srt=}')
    pred = truth[ [ srt[i,-j]  for j in range(1,k+1) for i in Range(truth)] ]
    true = np.hstack([truth for i in range(k)])
    return  precision_score(true, pred, average='micro')



import scanpy as sc
from scipy.sparse import csr_matrix
from sklearn.mixture import GaussianMixture
import pandas as pd
from lmz import *
import anndata as ad
import numpy as np
import random


####
# oldest datasets
#######
load = lambda f: [l for l in open(f,'r').read().split('\n') if len(l)>1]


def do_subsample(adata, subsample, seed = None):
    if not subsample:
        return adata

    if subsample <1:
        sc.pp.subsample(adata, fraction=subsample, n_obs=None, random_state=seed, copy=False)
    else:
        if adata.shape[0] < subsample:
            return adata
        sc.pp.subsample(adata, fraction=None, n_obs=subsample, random_state=seed, copy=False)
    return adata

def loadlabels(labels, ids):
    cellid_to_clusterid = {row.split(',')[0]:hash(row.split(',')[1]) for row in labels[1:]} #
    clusterid_to_nuclusterid = {item:clusterid for clusterid, item in enumerate(sorted(list(set(cellid_to_clusterid.values()))))}
    #print (clusterid_to_nuclusterid)
    return np.array([ clusterid_to_nuclusterid.get(cellid_to_clusterid.get(idd[:-2],-1),-1)  for idd in ids])

def filter(adata, cells='mito'):

    if cells == 'seurat':
        adata = adata[adata.obs['labels'] != -1,:] # -1 are unannotated cells, there are quite a lot of these here :)
    elif cells == 'mito':
        mito_genes = adata.var_names.str.startswith('MT-')
        adata.obs['percent_mito'] = np.sum(adata[:, mito_genes].X, axis=1).A1 / np.sum(adata.X, axis=1).A1
        adata = adata[adata.obs['percent_mito'] < 0.05, :]

    return adata

def load3k(cells: 'mito all seurat' ='mito', subsample=.15, seed = None, pathprefix = '..')-> 'anndata object':
    adata =  sc.read_10x_mtx(
    pathprefix+'/data/3k/hg19/',
    var_names='gene_symbols', cache=True)
    adata.obs['labels']= loadlabels(load( pathprefix+"/data/3k/pbmc.3k.labels"), load(pathprefix+"/data/3k/hg19/barcodes.tsv"))
    adata = filter(adata,cells)
    adata = do_subsample(adata, subsample,seed)
    return adata

def load6k(cells: 'mito all seurat' ='mito', subsample=.25, seed=None, pathprefix = '..')-> 'anndata object':
    adata =  sc.read_10x_mtx(
    pathprefix+'/data/6k/hg19/',

    var_names='gene_symbols', cache=True)

    adata.obs['labels']= loadlabels(load( pathprefix+"/data/6k/pbmc.6k.labels"), load( pathprefix+"/data/6k/hg19/barcodes.tsv"))

    adata = filter(adata,cells)
    adata = do_subsample(adata, subsample, seed)
    return adata

def loadpbmc(path=None, subsample=None, seed=None):
    adata = sc.read_10x_mtx( path,  var_names='gene_symbols', cache=True)
    adata = do_subsample(adata, subsample,seed)
    return adata

def loadGSM(path=None, subsample=None, seed=None, cellLabels=None, labelFile='theirLabels.csv'):
    adata = sc.read_10x_mtx( path,  var_names='gene_symbols', cache=True)
    if cellLabels:
        adata.obs['labels'] = pd.read_csv(path+'theirLabels.csv', usecols=["cluster"])['cluster'].to_numpy()
    adata = do_subsample(adata, subsample,seed)
    return adata

def load3k6k(subsample=False,seed=None, pathprefix = '..'):
    return load3k(subsample=subsample, seed=seed, pathprefix = pathprefix), load6k(subsample=subsample,seed=seed, pathprefix = pathprefix)

def loadp7de(subsample=False,pathprefix='..', seed=None):
    return loadpbmc('%s/data/p7d'%pathprefix ,subsample,seed), loadpbmc('%s/data/p7e'%pathprefix,subsample, seed)

def load4k8k(subsample=False,pathprefix='..',seed=None):
    return loadpbmc('%s/data/4k'% pathprefix,subsample, seed=seed), loadpbmc('%s/data/8k'%pathprefix,subsample,seed=seed)

def loadimmune(subsample=False, pathprefix='..',seed=None):
    return loadpbmc('%s/data/immune_stim/8'% pathprefix,subsample,seed=seed),\
           loadpbmc('%s/data/immune_stim/9'%pathprefix,subsample,seed=seed)


###
# grunreader
####

def loadgruen_single(path,subsample):
    mtx_path = path+".1.counts_raw.csv.gz"
    things = pd.read_csv(mtx_path, sep='\t').T
    adata = ad.AnnData(things)
    truthpath = path+".cell_assignments.csv.gz"
    truth  = pd.read_csv(truthpath, sep='\t')
    #adata.obs['true']  = list(truth['assigned_cluster'])
    adata.obs['true']  = list(truth['celltype'])
    do_subsample(adata, subsample)
    return adata

def loadgruen(subsample=False, pathprefix='..', methods=['human1','human2']):
    return [loadgruen_single('%s/data/punk/%s'% (pathprefix,method),subsample) for method in methods]





#######################3
# LOADING 100 DATASETS
######################
def get100names(path = '../data/100/data'):
    return ['Adipose_10xchromium_SRA710104-SRS3334253_9399', 'Adipose_10xchromium_SRA710104-SRS3334254_9107', 'Adipose_10xchromium_SRA710104-SRS3334255_9510', 'Bonemarrow_10xchromium_SRA779509-SRS3805245_6248', 'Bonemarrow_10xchromium_SRA779509-SRS3805246_6024', 'Bonemarrow_10xchromium_SRA779509-SRS3805247_7623', 'Bonemarrow_10xchromium_SRA779509-SRS3805248_6395', 'Bonemarrow_10xchromium_SRA779509-SRS3805255_7559', 'Bonemarrow_10xchromium_SRA779509-SRS3805258_5683', 'Bonemarrow_10xchromium_SRA779509-SRS3805262_6431', 'Bonemarrow_10xchromium_SRA779509-SRS3805266_6210', 'Colon_10xchromium_SRA703206-SRS3296611_4826', 'Colon_10xchromium_SRA703206-SRS3296612_6476', 'Colon_10xchromium_SRA728025-SRS3454422_2322', 'Colon_10xchromium_SRA728025-SRS3454423_2283', 'Colon_10xchromium_SRA728025-SRS3454424_1898', 'Colon_10xchromium_SRA728025-SRS3454425_3429', 'Colon_10xchromium_SRA728025-SRS3454426_3423', 'Colon_10xchromium_SRA728025-SRS3454427_3840', 'Colon_10xchromium_SRA728025-SRS3454428_5459', 'Colon_10xchromium_SRA728025-SRS3454430_4910', 'Cordblood_10xchromium_SRA769148-SRS3747193_2674', 'Cordblood_10xchromium_SRA769148-SRS3747194_2581', 'Cordblood_10xchromium_SRA769148-SRS3747195_2160', 'Cordblood_10xchromium_SRA769148-SRS3747196_2292', 'Cordblood_10xchromium_SRA769148-SRS3747197_2500', 'Cordblood_10xchromium_SRA769148-SRS3747198_3906', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322339_2914', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322341_3493', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322342_3372', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322343_3940', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322345_3718', 'Kaposissarcoma_10xchromium_SRA843432-SRS4322346_4479', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870733_2290', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870734_2038', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870735_1761', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870743_1470', 'Liver_10xchromium_SRA716608-SRS3391629_4190', 'Liver_10xchromium_SRA716608-SRS3391630_4719', 'Liver_10xchromium_SRA716608-SRS3391631_1425', 'Liver_10xchromium_SRA716608-SRS3391632_6158', 'Liver_10xchromium_SRA716608-SRS3391633_6806', 'Ovariantumor_10xchromium_SRA634975-SRS2724911_2547', 'Ovariantumor_10xchromium_SRA634975-SRS2724912_2139', 'Ovariantumor_10xchromium_SRA634975-SRS2724913_2262', 'Ovariantumor_10xchromium_SRA634975-SRS2724914_2094', 'Placenta_10xchromium_SRA782908-SRS3815597_3618', 'Placenta_10xchromium_SRA782908-SRS3815600_7205', 'Placenta_drop-seq_SRA782908-SRS3815594_391', 'Placenta_drop-seq_SRA782908-SRS3815595_392', 'Placenta_drop-seq_SRA782908-SRS3815596_292', 'Placenta_drop-seq_SRA782908-SRS3815598_301', 'Placenta_drop-seq_SRA782908-SRS3815599_2189', 'Placenta_drop-seq_SRA782908-SRS3815601_614', 'Placenta_drop-seq_SRA782908-SRS3815602_548', 'Placenta_drop-seq_SRA782908-SRS3815603_247', 'Prostate_10xchromium_SRA742961-SRS3565195_6101', 'Prostate_10xchromium_SRA742961-SRS3565196_7789', 'Prostate_10xchromium_SRA742961-SRS3565197_11468', 'Prostate_10xchromium_SRA742961-SRS3565198_7718', 'Prostate_10xchromium_SRA742961-SRS3565199_8069', 'Prostate_10xchromium_SRA742961-SRS3565201_7479', 'Prostate_10xchromium_SRA742961-SRS3565203_6453', 'Prostate_10xchromium_SRA742961-SRS3565206_6950', 'Prostate_10xchromium_SRA742961-SRS3565208_8538', 'Prostate_10xchromium_SRA742961-SRS3565211_6264', 'Tcells_10xchromium_SRA665712-SRS3034950_10979', 'Tcells_10xchromium_SRA665712-SRS3034951_14332', 'Tcells_10xchromium_SRA665712-SRS3034952_7788', 'Tcells_10xchromium_SRA665712-SRS3034953_12974', 'Tcells_10xchromium_SRA794656-SRS3937924_1878', 'Tcells_10xchromium_SRA794656-SRS3937926_1289', 'Tcells_10xchromium_SRA814476-SRS4073850_1975', 'Tcells_drop-seq_SRA867342-SRS4550171_2794', 'Tcells_drop-seq_SRA867342-SRS4550172_4068', 'Tcells_drop-seq_SRA867342-SRS4550173_2768', 'Testicle_10xchromium_SRA667709-SRS3065426_2777', 'Testicle_10xchromium_SRA667709-SRS3065427_3045', 'Testicle_10xchromium_SRA667709-SRS3065428_3007', 'Testicle_10xchromium_SRA667709-SRS3065429_3066', 'Testicle_10xchromium_SRA667709-SRS3065430_4020', 'Testicle_10xchromium_SRA667709-SRS3065431_3586', 'Testis_10xchromium_SRA645804-SRS2823404_4197', 'Testis_10xchromium_SRA645804-SRS2823405_3598', 'Testis_10xchromium_SRA645804-SRS2823406_3989', 'Testis_10xchromium_SRA645804-SRS2823407_4046', 'Testis_10xchromium_SRA645804-SRS2823408_4306', 'Testis_10xchromium_SRA645804-SRS2823409_4791', 'Testis_10xchromium_SRA645804-SRS2823410_4045', 'Testis_10xchromium_SRA645804-SRS2823411_2167', 'Testis_10xchromium_SRA645804-SRS2823412_5299', 'Testis_10xchromium_SRA645804-SRS3572594_4574', 'Tumor_10xchromium_SRA658915-SRS2944101_906', 'Tumor_10xchromium_SRA658915-SRS2944102_1229', 'Tumor_10xchromium_SRA658915-SRS2944103_865', 'Tumor_10xchromium_SRA658915-SRS2944104_1031', 'Tumor_10xchromium_SRA658915-SRS2944109_993', 'Tumor_10xchromium_SRA658915-SRS2944110_939', 'Tumor_10xchromium_SRA658915-SRS2944111_1197', 'Tumor_10xchromium_SRA658915-SRS2944112_1242']

def get41names():
    # see notebooks, the placenta dataset has somehow few cells
    # and there seems nothing wrong with adding testicle(not testies though those are all other the place) and cordblood ...
    return ['Bonemarrow_10xchromium_SRA779509-SRS3805245_6248', 'Bonemarrow_10xchromium_SRA779509-SRS3805246_6024', 'Bonemarrow_10xchromium_SRA779509-SRS3805247_7623', 'Bonemarrow_10xchromium_SRA779509-SRS3805248_6395', 'Bonemarrow_10xchromium_SRA779509-SRS3805255_7559', 'Bonemarrow_10xchromium_SRA779509-SRS3805258_5683', 'Bonemarrow_10xchromium_SRA779509-SRS3805262_6431', 'Bonemarrow_10xchromium_SRA779509-SRS3805266_6210', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870733_2290', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870734_2038', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870735_1761', 'Kidneyorganoids_drop-seq_SRA652805-SRS2870743_1470', 'Liver_10xchromium_SRA716608-SRS3391629_4190', 'Liver_10xchromium_SRA716608-SRS3391630_4719', 'Liver_10xchromium_SRA716608-SRS3391632_6158', 'Liver_10xchromium_SRA716608-SRS3391633_6806', 'Prostate_10xchromium_SRA742961-SRS3565195_6101', 'Prostate_10xchromium_SRA742961-SRS3565196_7789', 'Prostate_10xchromium_SRA742961-SRS3565197_11468', 'Prostate_10xchromium_SRA742961-SRS3565198_7718', 'Prostate_10xchromium_SRA742961-SRS3565199_8069', 'Prostate_10xchromium_SRA742961-SRS3565201_7479', 'Prostate_10xchromium_SRA742961-SRS3565203_6453', 'Prostate_10xchromium_SRA742961-SRS3565206_6950', 'Prostate_10xchromium_SRA742961-SRS3565208_8538', 'Prostate_10xchromium_SRA742961-SRS3565211_6264', 'Tcells_10xchromium_SRA665712-SRS3034950_10979', 'Tcells_10xchromium_SRA665712-SRS3034951_14332', 'Tcells_10xchromium_SRA665712-SRS3034952_7788', 'Tcells_10xchromium_SRA665712-SRS3034953_12974', 'Testicle_10xchromium_SRA667709-SRS3065427_3045', 'Testicle_10xchromium_SRA667709-SRS3065428_3007', 'Testicle_10xchromium_SRA667709-SRS3065429_3066', 'Testicle_10xchromium_SRA667709-SRS3065430_4020', 'Testicle_10xchromium_SRA667709-SRS3065431_3586', 'Cordblood_10xchromium_SRA769148-SRS3747193_2674', 'Cordblood_10xchromium_SRA769148-SRS3747194_2581', 'Cordblood_10xchromium_SRA769148-SRS3747195_2160', 'Cordblood_10xchromium_SRA769148-SRS3747196_2292', 'Cordblood_10xchromium_SRA769148-SRS3747197_2500', 'Cordblood_10xchromium_SRA769148-SRS3747198_3906']

def get100gz(item, path = '../data/100/data'):
    mtx_path = f"{path}/{item}.counts.gz"
    print("load path: ", mtx_path)
    things = pd.read_csv(mtx_path, sep='\t').T
    adata = ad.AnnData(things)
    adata.X=csr_matrix(adata.X)
    return adata
    #truthpath = path+".cell_assignments.csv.gz"
    #truth  = pd.read_csv(truthpath, sep='\t')
    #adata.obs['true']  = list(truth['assigned_cluster'])
    #adata.obs['true']  = list(truth['celltype'])

def load100(item, path='../data/100/data', seed= None, subsample=None, remove_unlabeled = False):
    adata =  ad.read_h5ad(f"{path}/{item}.h5")

    if remove_unlabeled:
        adata = adata[adata.obs['true']!=-1]

    i = adata.X.shape
    if subsample:
        try:
            sc.pp.subsample(adata, fraction=None, n_obs=subsample, random_state=seed, copy=False)
        except:
            raise Exception(f"COULD NOT SUBSAMPLE {subsample} items {path} {item}")
    #print(f"LOADING: SHAPE {i} (subsample)-> {adata.X.shape}")
    return adata

def load100addtruthAndWrite(adata,item, path='../data/100/data'):
    fname = f"{path}/{item}.cluster.txt"
    lol = open(fname,'r').readlines()
    barcode_cid={}
    for line in lol:
        bc,cl =  line.strip().split()
        barcode_cid[bc]= int(cl)
    #adata = ad.read_h5ad(fname)
    adata.obs['true'] = [barcode_cid.get(a,-1)  for a in adata.obs.index]
    fname = f"{path}/{item}.h5"
    adata.write(fname, compression='gzip')






from cellsaw.similarity.measures import cosine, jaccard, precision
import numpy as np
from ubergauss import tools as ut
from cellsaw.similarity import draw

def matrixmap(method, instances, repeats = 5):
    l = len(instances)
    res = np.zeros((l,l,repeats))
    for i,obj_i in enumerate(instances):
        for j,obj_j in enumerate(instances[i:]):
            r = [method(obj_i,obj_j,x) for x in range(repeats)]
            for x,val in enumerate(r):
                res[i,j,x] = val
                res[j,i,x] = val
    return res


def matrixmap_mp(method, instances, repeats = 5):
    # pool.maxtasksperchild
    l = len(instances)
    res = np.zeros((l,l,repeats))

    def func(stuff):
        a,b,seeds,i,j = stuff
        r = [ method(a,b,seed)  for seed in seeds ]
        return r,i,j

    tmptasks =([instances[i],instances[j],list(range(repeats)), i,j]
               for i in range(l) for j in range(i+1,l))

    for r,i,j in ut.xmap(func,tmptasks):
        for x,val in enumerate(r):
            res[i,j,x] = val
            res[j,i,x] = val

    return res


def getNeighbors(matrix, labels, k = 2 ):
    sm = ut.spacemap(labels)
    codes = getnn(matrix,k)
    res = [[sm.getitem[z] for z in a] for a in codes]
    return res

def getnn(m,n):
    np.fill_diagonal(m, np.NINF)
    srt=  np.argsort(m, axis= 1)
    return [ [i]+srt[i,-n:].tolist() for i in range(srt.shape[0]) ]






from scipy.optimize import linear_sum_assignment as lsa
from sklearn import neighbors as nbrs, metrics
import numpy as np
from sklearn.semi_supervised import LabelSpreading
from scipy.sparse.csgraph import dijkstra
from lmz import Map
from ubergauss import tools


def diffuse(mergething, labelslist, pid = 1, neighbors = 30):


    Xlist = mergething.projections[pid]

    #def diffuseandscore(y1, y2, Xlist, neighbors=30):
    assert Xlist[0].shape == Xlist[1].shape, 'not sure why i assert this... probably because i havent tested it'



    y1len = Xlist[0].shape[0]

    lp_model = LabelSpreading(kernel=lambda x, y: mykernel(y1len, neighbors, x, y),
                              alpha=.2,
                              max_iter=30)

    startlabels = np.hstack(labelslist)
    args = np.vstack(Map(tools.zehidense, Xlist)), startlabels

    try:
        lp_model.fit(*args)
    except:
        print(args[0].shape)
        print(args[1].shape)

    all_labels = lp_model.transduction_
    return all_labels[:y1len], all_labels[y1len:]








def hungmat(x1, x2):
        x= metrics.euclidean_distances(x1,x2)
        r = np.zeros_like(x)
        a,b = lsa(x)
        r[a,b] = 1
        r2 = np.zeros((x.shape[1], x.shape[0]))
        r2[b,a] = 1 # rorated :)
        return r,r2


def mykernel(x1len=False, neighbors = 3, X=None, _=None, return_graph = False):
    assert x1len, 'need to know how large the first dataset ist :)'
    '''
    X are the stacked projections[0] (normalized read matrices)
    since this is a kernel, we return a similarity matrix

    - we can split it by 2 to get the original projections
    - we do neighbors to get quadrant 2 and 4
    - we do hungarian to do quadrants 1 and 3
    - we do dijkstra to get a complete distance matrix

    '''
    x1,x2 = np.split(X,[x1len])
    q2 = nbrs.kneighbors_graph(x1,neighbors).todense()
    q4 = nbrs.kneighbors_graph(x2,neighbors).todense()
    q1,q3 = hungmat(x1,x2)

    graph = np.hstack((np.vstack((q2,q3)),np.vstack((q1,q4))))


    connect = dijkstra(graph,unweighted = True, directed = False)

    if return_graph:
        return graph, connect
    distances = -connect # invert
    distances -= distances.min() # longest = 4
    distances /= distances.max() # between 0 and 1 :)
    distances[distances < np.median(np.unique(distances))] = 0

    return np.power(distances,2)








from cellsaw.load.loadadata import get41names, get100names, load100
from cellsaw.load.preprocess import annotate_genescores

def easyLoad100(name, path = None, remove_unlabeled = False, mingenes= 200,subsample = None,
                preprocessingmethod = 'natto', donormalize= True,
                plot=False,quiet=True, nattoargs=  {'mean': (0.015, 4), 'bins': (.25, 1)}):
    adata = load100(name, path=path, remove_unlabeled=remove_unlabeled, subsample = subsample)
    gs = annotate_genescores(adata, mingenes=mingenes, selector=preprocessingmethod, donormalize= donormalize, nattoargs= nattoargs, plot=plot, quiet=quiet)
    return gs


import matplotlib.pyplot as plt
from lmz import *
import sklearn
import seaborn as sns
from anndata import AnnData
import numpy as np
import pandas as pd
import scanpy as sc
from scipy.sparse import csr_matrix
from lmz import *

def annotate_genescores(adata, selector='natto',
                        donormalize=True,
                        nattoargs = {'mean':(0.015, 4),'bins':(.25, 1)},
                        mingenes = 200,
                        quiet = False,
                        plot=False):

    incommingshape= adata.X.shape
    sc.pp.filter_cells(adata, min_genes=mingenes, inplace=True)
    okgenes = sc.pp.filter_genes(adata, min_counts=3, inplace=False)[0]
    if donormalize:
        sc.pp.normalize_total(adata, 1e4)
        sc.pp.log1p(adata)

    adata2 = adata.copy()
    adata = adata[:,okgenes]
    if selector == 'preselected':
        self.preselected_genes = self.data[0].preselected_genes

    if selector == 'natto':
        genes, scores = getgenes_natto(adata, 1000, plot=plot, **nattoargs)

    elif selector == 'preselected':
        genes = [True if gene in self.preselected_genes else False for gene in adata.var_names]
        scores = genes.as_type(int)

    else:
        hvg_df = sc.pp.highly_variable_genes(adata, n_top_genes=1000, flavor=selector, inplace=False)
        genes = np.array(hvg_df['highly_variable'])
        if selector == 'seurat_v3':
            ### Best used for raw_count data
            scores = np.array(hvg_df['variances_norm'].fillna(0))
        else:
            scores = np.array(hvg_df['dispersions_norm'].fillna(0))


    #fullscores = np.zeros(adata2.X.shape[1])
    fullscores = np.full(adata2.X.shape[1],np.NINF,np.float)
    fullscores[okgenes==1] = scores
    adata2.varm["scores"]=  fullscores
    adata2.varm['genes'] = okgenes
    #adata.varm["genes"] = genes ... lets decide later if we need this
    if not quiet:
        print(f"{incommingshape=}  => {adata.X.shape}")
    return adata2







####
# ft select
###
def transform( means, var,plot, stepsize=.5, ran=3, minbin=0 ):
    x = np.arange(minbin * stepsize, ran, stepsize) #-> .5,1,1.5,2,2.5 ...

    #items = [(m, v) for m, v in zip(means, var)] #-> items = list(zip(means,var))
    items = Zip(means,var)

    boxes = [[i[1] for i in items if r < i[0] < r + (stepsize)] for r in x]


    ystdx = [(np.median(box),np.std(box),xcoo) for box,xcoo in zip(boxes,x) if box]

    #y = np.array([np.median(st) for st in boxes])
    #y_std = np.array([np.std(st) for st in boxes])
    y,y_std,x = map(np.array,Transpose(ystdx))

    x = x + (stepsize / 2)
    # draw regression points
    if plot:
        plt.scatter(x, y, label='Mean of bins', color='k')

    nonan = np.isfinite(y)
    x = x[nonan]
    y = y[nonan]
    y_std = y_std[nonan]
    x = x.reshape(-1, 1)
    return x, y, y_std


def get_expected_values(x, y, x_all, firstbinchoice = max):
    mod = sklearn.linear_model.HuberRegressor()
    mod.fit(x, y)
    res = mod.predict(x_all.reshape(-1, 1))
    firstbin = y[0]
    firstbin_esti = mod.predict([x[0]])
    res[x_all < x[0]] = firstbinchoice(firstbin, firstbin_esti)
    return res


def getgenes_natto(adata, selectgenes,
        mean=(.015,4),
        bins=(.25,1),
        plot=True):

    matrix= adata.to_df().to_numpy()
    a = np.expm1(matrix)
    var = np.var(a, axis=0)
    meanex = np.mean(a, axis=0)
    with np.errstate(divide='ignore', invalid='ignore'):
        # error will happen but we catch it later..
        disp = var / meanex
    Y = np.log(disp)
    X = np.log1p(meanex)


    #plt.scatter(X,Y)
    #plt.show()

    mask = np.array([not np.isnan(y) and me > mean[0] and me < mean[1] for y, me in zip(disp, X)])
    if plot:
        sns.set_style("whitegrid")
        plt.figure(figsize=(11, 4))
        plt.suptitle(f"gene selection", size=20, y=1.07)
        ax = plt.subplot(121)
        plt.scatter(X[mask], Y[mask], alpha=.2, s=3, label='all genes')

    x_bin, y_bin, ystd_bin = transform(X[mask].reshape(-1, 1),
                                            Y[mask],plot,
                                            stepsize=bins[0],
                                            ran=mean[1],
                                            minbin=bins[1] )


    y_predicted = get_expected_values(x_bin, y_bin, X[mask])
    std_predicted = get_expected_values(x_bin, ystd_bin, X[mask])
    Y[mask] -= y_predicted
    Y[mask] /= std_predicted

    srt = np.argsort(Y[mask])
    accept = np.full(Y[mask].shape, False)
    accept[srt[-selectgenes:]] = True

    if plot:
        srt = np.argsort(X[mask])
        plt.plot(X[mask][srt], y_predicted[srt], color='k', label='regression')
        plt.plot(X[mask][srt], std_predicted[srt], color='g', label='regression of std')
        plt.scatter(x_bin, ystd_bin, alpha=.4, label='Std of bins', color='g')
        plt.legend(bbox_to_anchor=(.6, -.2))
        plt.title("dispersion of genes")
        plt.xlabel('log mean expression')
        plt.ylabel('dispursion')
        ax = plt.subplot(122)
        plt.scatter(X[mask], Y[mask], alpha=.2, s=3, label='all genes')
        g = X[mask]
        d = Y[mask]
        plt.scatter(g[accept], d[accept], alpha=.3, s=3, color='r', label='selected genes')
        plt.legend(bbox_to_anchor=(.6, -.2))
        plt.title("normalized dispersion of genes")
        plt.xlabel('log mean expression')
        plt.ylabel('dispursion')
        plt.show()

        print(f"ft selected:{sum(accept)}")

    raw = np.zeros(len(mask))
    raw[mask] = Y[mask]

    mask[mask] = np.array(accept)
    return mask, raw



import cellsaw.merge.mergehelpers  as mergehelpers
import numpy as np
from cellsaw.merge.diffusion import diffuse
from cellsaw.merge import draw

class mergeutils:
    def getlabels(self,masked = []):
        f = lambda i,e: np.full_like(e.obs['true'],-1)  if i in masked else e.obs['true']
        return [ f(i,e) for i,e in enumerate(self.data)]


class merge(mergeutils):
    def __init__(self, adatas, selectgenes = 800, make_even = True, pca = 20, umaps = [2],
            joint_space = False,
            sortfield = 0,
            titles = "ABCDEFGHIJKIJ"):

        shapesbevorepp= [a.X.shape for a in adatas]
        self.genescores = [a.varm['scores'] for a in adatas]
        self.geneab = [a.varm['genes'] for a in adatas]
        self.data  = mergehelpers.unioncut(self.genescores, selectgenes, adatas)

        # geneab = np.all(np.array(self.geneab), axis=0)
        # for i, d in enumerate(self.data):
        #     data[i] = d[:, geneab]

        if make_even:
            self.data = mergehelpers.make_even(self.data)


        print("preprocess:", end= '')
        for a,b in zip(shapesbevorepp, self.data):
            print(f"{a} -> {b.shape}")

        ######################
        self.titles = titles

         # do dimred
        self.projections = [[ d.X for d in self.data]]+mergehelpers.dimension_reduction(self.data,
                                                                False, # scale (will still scale if pca)
                                                                False, # will be ignored anyway
                                                                PCA=pca,
                                                                umaps=umaps,
                                                                joint_space=joint_space)

        if pca:
            self.PCA = self.projections[1]

        if sortfield >=0:
            self.sort_cells(sortfield)

        if umaps:
            for x,d in zip(umaps,self.projections[int(pca>0)+1:]):
                self.__dict__[f"d{x}"] = d


    def sort_cells(self,projection_id = 0):
        # loop over data sets
        self.hungdist =[]
        for i in range(len(self.data)-1):
            hung, dist = self.hungarian(projection_id,i,i+1)
            self.hungdist.append(dist)
            #self.data[i+1] = self.data[i+1][hung[1]]
            #self.data[i+1].X = self.data[i+1].X[hung[1]]
            for x in range(len(self.projections)):
                self.projections[x][i+1] = self.projections[x][i+1][hung[1]]
                if x == 0:
                    self.data[i+1]= self.data[i+1][hung[1]]
    #
    def hungarian(self,data_fld,data_id, data_id2):
            hung, dist = mergehelpers.hungarian(self.projections[data_fld][data_id],self.projections[data_fld][data_id2])
            return hung, dist[hung]





import matplotlib.pyplot as plt
from ubergauss import tools
import numpy as np
from umap import UMAP
import seaborn as sns
from scipy.optimize import linear_sum_assignment as lsa
from lmz import *


col = plt.cm.get_cmap('tab20').colors
col = col+col+col+ ((0,0,0),)



def plot(merge, labels, plotsperline =3, grad=False):


    # make a tinyumap with the right dimensions
    X = merge.d2
    rows = ((len(X)-1)//plotsperline)+1
    columns  = plotsperline if  len(X) > plotsperline else len(X)
    d = tinyUmap(dim = (rows, columns))  # default is a row



    # set same limit for all the plots
    concatX  = np.vstack(X)
    xmin,ymin = concatX.min(axis = 0)
    xmax,ymax = concatX.max(axis = 0)
    xdiff = np.abs(xmax - xmin)
    ydiff = np.abs(ymax - ymin)

    for x,y in zip(X,labels):

        if not grad:
            d.draw(x,y, title=None)

            plt.xlim(xmin - 0.1 * xdiff, xmax + 0.1 * xdiff)
            plt.ylim(ymin - 0.1 * ydiff, ymax + 0.1 * ydiff)

            #plt.legend(markerscale=1.5, fontsize='small', ncol=int(len(X) * 2.5), bbox_to_anchor=(1, -.12))
        else:
            plt.scatter(x[:,0], x[:,1], c=y, s=1)

    plt.show()



def tinyumap(X,Y,
        title="No title",
        title_size=10,
        acc : "y:str_description"={},
        markerscale=4,
        getmarker = lambda cla: {"marker":'o'},
        col=col,
        label=None,
        alpha = None,
        legend = False,
        size=None):
    plt.title(title, size=title_size)
    Y=np.array(Y)
    size=  max( int(4000/Y.shape[0]), 1) if not size else size
    embed = X
    plt.gca().axes.yaxis.set_ticklabels([])
    plt.gca().axes.xaxis.set_ticklabels([])
    for cla in np.unique(Y):
        plt.scatter(embed[Y==cla, 0],
                    embed[Y==cla, 1],
                    color= col[cla],
                    s=size,
                    edgecolors = 'none',
                    alpha = alpha,
                    label=str(cla), **getmarker(cla)) #str(cla)+" "+acc.get(cla,''),**getmarker(col[cla]))
    #plt.axis('off')
    #plt.xlabel('UMAP 2')
    #plt.ylabel('UMAP 1')
    if legend:
        plt.legend(markerscale=2,ncol=2,bbox_to_anchor=(1, -.12) )


class tinyUmap():

    def __init__(self, dim=(3,3), size= 2):
        figs = (size*dim[1], size*dim[0])

        plt.figure( figsize=figs, dpi=300)
        self.i =0
        self.dim = dim

    def next(self):
        self.i= self.i+1
        plt.subplot(*self.dim,self.i)

    def draw(self, *a, **b):
        self.next()
        tinyumap(*a,**b)


def mkconfusion(y1,y2):
    assert len(y1)== len(y2)
    sm2 = tools.spacemap(np.unique(y2))
    sm1 = tools.spacemap(np.unique(y1))
    res = np.zeros((sm1.len, sm2.len))
    for a,b in zip(y1,y2):
        res[sm1.getint[a], sm2.getint[b]]+=1
    return res, sm1, sm2


def confuse(y1,y2, norm = True, draw= False):
    '''
    the task here is to draw a confision matrix
    returns translated labels for y2
    '''
    cm,sm1,sm2 = mkconfusion(y1,y2)


    #print(cm.shape, len(np.unique(y1)),  len(np.unique(y2)))
    #print(np.unique(y1),  np.unique(y2))
    # sns.heatmap(cm,xticklabels=np.unique(y2),yticklabels=np.unique(y1), annot=True, linewidths=.5,cmap="YlGnBu" , square=True)
    # plt.show()
    # plt.close()

    itemCount2 = dict(zip(*np.unique(y2,return_counts = True)))
    itemCount1 = dict(zip(*np.unique(y1,return_counts = True)))


    if norm:
        cm = cm.astype(float)
        for x in Range(itemCount2):
            for y in Range(itemCount1):
                res = (2*cm[y,x]) / (itemCount2[sm2.getitem[x]]+itemCount1[sm1.getitem[y]])
                #print(cm[y,x])
                #print(xcounts[x],ycounts[y])
                cm[y,x]  = res

    #sns.heatmap(cm,xticklabels=np.unique(y2),yticklabels=np.unique(y1), annot=False, linewidths=.5,cmap="YlGnBu" , square=True)
    # plt.title('raw data')
    #plt.show()
    # plt.close()

    names1, new_order2 = lsa(cm, maximize= True)
    #print("hung matches:",names1,new_order2, itemCount2.keys())

    if draw:
        #new_order = np.argsort(new_order2)
        #cm[names1] = cm[new_order2]

        ylab = np.unique(y1)
        for n1,n2 in zip(names1, new_order2):
            cm[[n1, n2]] = cm[[n2, n1]]
            ylab[[n1,n2]]=ylab[[n2,n1]]


        sns.heatmap(cm,xticklabels=np.unique(y2),yticklabels=ylab, annot=False, linewidths=.5,cmap="YlGnBu" , square=True)
        plt.xlabel('Clusters data set 2')
        plt.ylabel('Clusters data set 1')

    g = dict(zip(new_order2, names1))
    return [ sm1.getitem[g[sm2.getint[y]]]  for y in y2]


def confuse2(labels):
    '''
    does the confusionmatrix twice, normalized and unnormalized, and plots!
    '''
    y1,y2 =  labels
    f=plt.figure(figsize=(16,8))
    ax=plt.subplot(121,title ='absolute hits')
    confuse(y1,y2,norm=False, draw= True)
    ax=plt.subplot(122,title ='relative hits (2x hits / sumOfLabels)')
    confuse(y1,y2,norm=True, draw= True)
    plt.tight_layout()
